/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Macroses to define a mediator for a hardware model
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/errors.hpp>
#include <hw/model.hpp>

#include <sstream>

////////////////////////////////////////////////////////////////////////////////
// Declaration of Mediator (Adapter) Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MEDIA
 * Declares a mediator (adapter) of a hardware model.
 */
#define CPPTESK_MEDIA(UserMedia, UserModel) \
    struct UserMedia: public UserModel

/** @def CPPTESK_SIGNAL_MEDIA
 * Declares a signal mediator (adapter) of a hardware model.
 */
#define CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs) \
    struct UserMedia: public UserModel, public ::cpptesk::hw::IOSignals<Inputs, Outputs>

/** @def CPPTESK_ADAPTER
 * Alias of CPPTESK_MEDIA
 */
#define CPPTESK_ADAPTER(UserMedia, UserModel) CPPTESK_MEDIA(UserMedia, UserModel)

/** @def CPPTESK_SIGNAL_ADAPTER
 * Alias of CPPTESK_SIGNAL_MEDIA
 */
#define CPPTESK_SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs) CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)

#ifdef HW_SHORT_NAMES
    #define MEDIA(UserMedia, UserModel)                           CPPTESK_MEDIA(UserMedia, UserModel)
    #define ADAPTER(UserMedia, UserModel)                         CPPTESK_ADAPTER(UserMedia, UserModel)
    #define SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)   CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)
    #define SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs) CPPTESK_SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs)
#endif // HW_SHORT_NAMES

////////////////////////////////////////////////////////////////////////////////
// Registration of Interface Adapters
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_SET_IFACE_ADAPTER
 * Registers an interface adapter (serializer or deserializer).
 */
#define CPPTESK_SET_IFACE_ADAPTER(iface, full_method_name) \
    registerInterfaceAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name)

/** @def CPPTESK_SET_INPUT_ADAPTER
 * Registers an input interface adapter (aka serializer).
 */
#define CPPTESK_SET_INPUT_ADAPTER(iface, full_method_name) \
    registerInputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name)


#if CPPTESKHW_CHECK_VERSION(111010)
#define CPPTESK_SET_OUTPUT_ADAPTER(iface, full_method_name) \
    do { \
        registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name); \
        CPPTESK_CALL_PARALLEL(listener, iface, iface##_message); \
    } while(false)
#define CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface, full_method_name) \
    do { \
        registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)
#define CPPTESK_SET_OUTPUT_ADAPTER_USER_LISTENER(iface, full_method_name, full_listener_name) \
    do { \
        registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name); \
        CPPTESK_CALL_LOOP_EX(full_listener_name,  iface, ::cpptesk::hw::HIGHEST_PRIORITY); \
    } while(false)
#else

/**e@def CPPTESK_SET_OUTPUT_ADAPTER
 * Registers an output interface adapter (aka deserializer).
 */
#define CPPTESK_SET_OUTPUT_ADAPTER(iface, full_method_name) \
    registerOutputAdapter(iface, (model_t::process_t)&full_method_name, #full_method_name)

#endif


/** @def CPPTESK_SET_INOUT_ADAPTER
 * Alias of CPPTESK_SET_IFACE_ADAPTER
 */
#define CPPTESK_SET_INOUT_ADAPTER(iface, full_method_name) \
    CPPTESK_SET_IFACE_ADAPTER(iface, full_method_name)

////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_SET_IFACE_ADAPTER_OF
 * Registers an interface adapter for an inner unit's interface.
 */
#define CPPTESK_SET_IFACE_ADAPTER_OF(UnitClass, unit, iface, full_method_name) \
    do { \
        unit.registerInterfaceAdapter(iface, reinterpret_cast<UnitClass::model_t&>(*this), \
            (UnitClass::model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)

/** @def CPPTESK_SET_INPUT_ADAPTER_OF
 * Registers an input interface adapter for an inner unit's interface.
 */
#define CPPTESK_SET_INPUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name) \
    do { \
        unit.registerInputAdapter(iface, reinterpret_cast<UnitClass::model_t&>(*this), \
            (UnitClass::model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)

/** @def CPPTESK_SET_OUTPUT_ADAPTER_OF
 * Registers an output interface adapter for an inner unit's interface.
 */
#define CPPTESK_SET_OUTPUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name) \
    do { \
        unit.registerOutputAdapter(iface, reinterpret_cast<UnitClass::model_t&>(*this), \
            (UnitClass::model_t::process_t)&full_method_name, #full_method_name); \
    } while(false)

/** @def CPPTESK_SET_INOUT_ADAPTER_OF
 * Alias of CPPTESK_SET_IFACE_ADAPTER_OF
 */
#define CPPTESK_SET_INOUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name) \
     CPPTESK_SET_INNER_IFACE_ADAPTER(UnitClass, unit, iface, full_method_name)

#ifdef HW_SHORT_NAMES
    #define SET_IFACE_ADAPTER(iface, full_method_name)                          CPPTESK_SET_IFACE_ADAPTER (iface, full_method_name)
    #define SET_INPUT_ADAPTER(iface, full_method_name)                          CPPTESK_SET_INPUT_ADAPTER (iface, full_method_name)
#if CPPTESKHW_CHECK_VERSION(111010)
    #define SET_OUTPUT_ADAPTER(iface, full_method_name)                         CPPTESK_SET_OUTPUT_ADAPTER(iface, full_method_name)
    #define SET_OUTPUT_ADAPTER_NO_LISTENER(iface, full_method_name)             CPPTESK_SET_OUTPUT_ADAPTER_NO_LISTENER(iface, full_method_name)
    #define SET_OUTPUT_ADAPTER_USER_LISTENER(iface, full_method_name, listener) CPPTESK_SET_OUTPUT_ADAPTER_USER_LISTENER(iface, full_method_name, listener)
#else
    #define SET_OUTPUT_ADAPTER(iface, full_method_name)                         CPPTESK_SET_OUTPUT_ADAPTER(iface, full_method_name)
#endif
    #define SET_INOUT_ADAPTER(iface, full_method_name)                          CPPTESK_SET_INOUT_ADAPTER (iface, full_method_name)
    #define SET_IFACE_ADAPTER_OF(UnitClass, unit, iface, full_method_name)      CPPTESK_SET_IFACE_ADAPTER_OF (UnitClass, unit, iface, full_method_name)
    #define SET_INPUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name)      CPPTESK_SET_INPUT_ADAPTER_OF (UnitClass, unit, iface, full_method_name)
    #define SET_OUTPUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name)     CPPTESK_SET_OUTPUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name)
    #define SET_INOUT_ADAPTER_OF(UnitClass, unit, iface, full_method_name)      CPPTESK_SET_INOUT_ADAPTER_OF (UnitClass, unit, iface, full_method_name)
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Reaction Arbitration
///////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_DETECT_REACTION_LABEL $CPPTESK_WAIT_REACTION_REACTION

#define __CPPTESK_CHECK_DETECT_REACTION() \
    do { \
        if(false) goto __CPPTESK_DETECT_REACTION_LABEL; \
    } while(false)

/**
 * @def CPPTESK_DETECT_REACTION
 * Detects a reaction by using a single-cycle condition but do not take it.
 * It is used when a detector is multi-cycle.
 */
#define CPPTESK_DETECT_REACTION(cond) \
    do { \
        if(CPPTESK_GET_PROCESS().listener) { \
            CPPTESK_WAIT(cond); \
	    if(CPPTESK_GET_IFACE().isModelBased()) \
                CPPTESK_CALL_PARALLEL_EX(model_t::singleListener, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE(), START_DONT_CLONE_MSG | START_IMMEDIATELY | ::cpptesk::hw::HIGHEST_PRIORITY); \
        } \
        else if(!CPPTESK_GET_MESSAGE().isDetected()) { \
            CPPTESK_WAIT_TIMEOUT(CPPTESK_GET_IFACE().isReady() || CPPTESK_GET_IFACE().getOwner() == &CPPTESK_GET_PROCESS(), getReactionTimeout()); \
            CPPTESK_GET_IFACE().appendRequest(CPPTESK_GET_PROCESS(), CPPTESK_GET_MESSAGE()); \
            /* Only one CPPTESK_DETECT_REACTION() (CPPTESK_WAIT_REACTION()) is allowed within one deserializer */ \
            __CPPTESK_DETECT_REACTION_LABEL: \
            CPPTESK_WAIT_TIMEOUT((cond) && (CPPTESK_GET_IFACE().isModelBased() ? \
	                         CPPTESK_GET_IFACE().chooseRequest(CPPTESK_GET_PROCESS(), CPPTESK_GET_MESSAGE()) : \
                                 CPPTESK_GET_IFACE().preChooseRequest(CPPTESK_GET_PROCESS(), CPPTESK_GET_MESSAGE())),  getReactionTimeout()); \
        } \
    } while(false)

/**
 * @def CPPTESK_TAKE_REACTION
 * Takes a reaction (says to the interface that the reaction is taken).
 */
#define CPPTESK_TAKE_REACTION() \
    do { \
        if(CPPTESK_GET_PROCESS().listener) { \
	} \
        else if(!CPPTESK_GET_MESSAGE().isDetected()) { \
            if(CPPTESK_GET_PROCESS().timeout) { \
                CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_WARN, "Time limit in WAIT_TIMEOUT() has been reached"); \
            } \
            /* Error if a reaction has been taken by someone else in case of modelBased arbiters */ \
            if(CPPTESK_GET_IFACE().isModelBased()) { \
                if(!CPPTESK_GET_IFACE().takeReaction()) { \
                    CPPTESK_ASSERTION(false, "Two output adapters have started for the same reaction on iface=" << \
                        CPPTESK_GET_IFACE().getName() << " (use SET_ARBITER() to set a reaction arbiter)"); \
                } \
            } \
        } \
    } while(false)

/** @def CPPTESK_WAIT_REACTION
 * Detects a reaction by using a single-cycle condition and takes it.
 */
#define CPPTESK_WAIT_REACTION(cond) \
    do { \
        CPPTESK_DETECT_REACTION(cond); \
        /* Reaction is taken only if time limit is not reached */ \
        if(!CPPTESK_GET_PROCESS().timeout) { \
            CPPTESK_TAKE_REACTION(); \
        } \
    } while(false)

/** @def CPPTESK_NEXT_REACTION
 * Finishes processing of a reaction and switches to the next one.
 */
#define CPPTESK_NEXT_REACTION() \
    do { \
        if(CPPTESK_GET_PROCESS().listener) { \
            if(!CPPTESK_GET_IFACE().isModelBased()) \
                CPPTESK_CALL_PARALLEL_EX(model_t::singleListener, CPPTESK_GET_IFACE(), CPPTESK_GET_MESSAGE(), START_DONT_CLONE_MSG | START_IMMEDIATELY | ::cpptesk::hw::HIGHEST_PRIORITY); \
            CPPTESK_CYCLE(); \
	} \
        else if(!CPPTESK_GET_MESSAGE().isDetected()) { \
            /* Check that CPPTESK_DETECT_REACTION() (CPPTESK_WAIT_REACTION()) exists */ \
            __CPPTESK_CHECK_DETECT_REACTION(); \
            if(!CPPTESK_GET_IFACE().isModelBased()) { \
                if(!CPPTESK_GET_IFACE().chooseRequest(CPPTESK_GET_PROCESS(), CPPTESK_GET_MESSAGE())) { \
                    CPPTESK_CYCLE(); \
                    goto __CPPTESK_DETECT_REACTION_LABEL; \
                } \
                if(!CPPTESK_GET_IFACE().takeReaction()) { \
                    CPPTESK_ASSERTION(false, "Two adaptive output adapters have started for the same reaction on iface=" << \
                        CPPTESK_GET_IFACE().getName() << " (data in messages were the same; try cascade arbiter)"); \
                } \
            } \
            ::cpptesk::hw::priority_t __cpptesk_priority = (::cpptesk::hw::priority_t)(CPPTESK_GET_PRIORITY()); \
	    /* HIGHEST_PRIORITY -1 is used due to the most highest priority of listener: it should check the reaction has been taken and then such a flag should be unset */ \
            CPPTESK_SET_PRIORITY((::cpptesk::hw::priority_t)(::cpptesk::hw::HIGHEST_PRIORITY - 1)); \
            CPPTESK_CYCLE(); \
            CPPTESK_SET_PRIORITY(__cpptesk_priority); \
            CPPTESK_GET_IFACE().removeRequest(CPPTESK_GET_PROCESS(), CPPTESK_GET_MESSAGE()); \
        } \
    } while(false)

#ifdef HW_SHORT_NAMES
    #define DETECT_REACTION(cond) CPPTESK_DETECT_REACTION(cond)
    #define TAKE_REACTION()       CPPTESK_TAKE_REACTION()
    #define WAIT_REACTION(cond)   CPPTESK_WAIT_REACTION(cond)
    #define NEXT_REACTION()       CPPTESK_NEXT_REACTION()
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Declaration of Listeners
///////////////////////////////////////////////////////////////////////////////
/*
Use Case of User-Defined Listener
DECLARE_LISTENER(listen_iface1)
{
    START_LISTENING(iface1);
    LISTENING() {
        WAIT(cond);
	CYCLE();
	if(!REACTION_TAKEN())
	    DESERIALIZE_UNEXPECTED_REACTION();
    }
    STOP_LISTENING();
}
*/
#if CPPTESKHW_CHECK_VERSION(111010)

#define CPPTESK_DECLARE_LISTENER(listener_name) \
    CPPTESK_DECLARE_PROCESS(listener_name)

#define CPPTESK_DEFINE_LISTENER(listener_name) \
    CPPTESK_DEFINE_PROCESS(listener_name)

#define CPPTESK_START_LISTENER(msgtype) \
    msgtype impl_msg; \
    ::cpptesk::hw::cycle_t start_time, stop_time; \
    ::cpptesk::hw::pid_t pid = CPPTESK_INCORRECT_PID; \
    CPPTESK_UNUSED_ARG(CPPTESK_GET_PROCESS()); \
    CPPTESK_UNUSED_ARG(CPPTESK_GET_MESSAGE()); \
    CPPTESK_START_PROCESS()

#define CPPTESK_LISTENING() \
    while(true)

#define CPPTESK_CYCLE_BEFORE_CHECKING() \
    adapters_map_t::iterator i = _adapters.find(&iface); \
    start_time = time(); \
    if(i != _adapters.end()) { \
        impl_msg.setDetected(true); \
        impl_msg.invalidateFields(); \
        pid = i->second->call_parallel(CPPTESK_GET_IFACE(), impl_msg); \
    } \
    CPPTESK_CYCLE()

#define CPPTESK_IS_REACTION_TAKEN() \
    CPPTESK_GET_IFACE().isReactionTaken()

#define CPPTESK_DESERIALIZE_UNEXPECTED_REACTIONS() \
        do { \
            std::stringstream buffer; \
            CPPTESK_WAIT_PROCESS_FINISHED(pid); \
            stop_time = time(); \
            buffer << "Unexpected reaction: " << impl_msg; \
            ::cpptesk::hw::Reaction reaction(::cpptesk::hw::UNEXPECTED_REACTION, CPPTESK_GET_IFACE(), impl_msg, start_time, stop_time, buffer.str()); \
            CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_ERROR, reaction); \
            registerReaction(reaction); \
        } while(false)

#define CPPTESK_STOP_LISTENER() \
    CPPTESK_STOP_PROCESS()

#ifdef HW_SHORT_NAMES
#define DECLARE_LISTENER(listener_name) CPPTESK_DECLARE_LISTENER(listener_name)
#define DEFINE_LISTENER(listener_name) CPPTESK_DEFINE_LISTENER(listener_name)
#define START_LISTENER(iface) CPPTESK_START_LISTENER(iface)
#define LISTENING() CPPTESK_LISTENING()
#define CYCLE_BEFORE_CHECKING() CPPTESK_CYCLE_BEFORE_CHECKING()
#define IS_REACTION_TAKEN() CPPTESK_IS_REACTION_TAKEN()
#define DESERIALIZE_UNEXPECTED_REACTIONS() CPPTESK_DESERIALIZE_UNEXPECTED_REACTIONS()
#define STOP_LISTENER() CPPTESK_STOP_LISTENER()
#endif

#endif //CPPTESKHW_CHECK_VERSION(111010)

#ifdef CPPTESK_DESERIALIZE_UNEXPECTED_REACTION
#define CPPTESK_DEFINE_BASIC_OUTPUT_LISTENER(listener_name, output_iface, message_type, condition) \
    CPPTESK_DECLARE_PROCESS(listener_name) { \
        message_type impl_msg; \
        ::cpptesk::hw::cycle_t start_time, stop_time; \
        ::cpptesk::hw::pid_t pid = CPPTESK_INCORRECT_PID; \
        CPPTESK_UNUSED_ARG(CPPTESK_GET_PROCESS()); \
        CPPTESK_UNUSED_ARG(CPPTESK_GET_MESSAGE()); \
        \
        CPPTESK_START_PROCESS(); \
        while(true) { \
            CPPTESK_WAIT(condition); \
            adapters_map_t::iterator i = _adapters.find(&iface); \
            \
            start_time = time(); \
            \
            if(i != _adapters.end()) { \
                impl_msg.setDetected(true); \
                impl_msg.invalidateFields(); \
                pid = i->second->call_parallel(CPPTESK_GET_IFACE(), impl_msg); \
            } \
            CPPTESK_CYCLE(); \
            \
            if(!CPPTESK_GET_IFACE().isReactionTaken()) { \
                std::stringstream buffer; \
                \
                CPPTESK_WAIT_PROCESS_FINISHED(pid); \
                stop_time = time(); \
                \
                buffer << "Unexpected reaction: " << impl_msg; \
                \
                ::cpptesk::hw::Reaction reaction(::cpptesk::hw::UNEXPECTED_REACTION, \
                    CPPTESK_GET_IFACE(), impl_msg, start_time, stop_time, buffer.str()); \
                CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_ERROR, reaction); \
                registerReaction(reaction); \
            } \
        } \
        CPPTESK_STOP_PROCESS(); \
    }
#else
#define CPPTESK_DEFINE_BASIC_OUTPUT_LISTENER(listener_name, output_iface, condition) \
    CPPTESK_DECLARE_PROCESS(listener_name) { \
        CPPTESK_UNUSED_ARG(CPPTESK_GET_PROCESS()); \
        CPPTESK_UNUSED_ARG(CPPTESK_GET_MESSAGE()); \
        CPPTESK_START_PROCESS(); \
        while(true) { \
            CPPTESK_WAIT(condition); \
            CPPTESK_CYCLE(); \
            \
            if(!iface.isReactionTaken()) { \
                ::cpptesk::hw::Reaction reaction(::cpptesk::hw::UNEXPECTED_REACTION, \
                    CPPTESK_GET_IFACE(), time(), time(), "Unexpected reaction"); \
                CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_ERROR, reaction); \
                registerReaction(reaction); \
             } \
          } \
          CPPTESK_STOP_PROCESS(); \
      }
#endif

#define CPPTESK_CALL_OUTPUT_LISTENER(full_listener_name, output_iface) \
    CPPTESK_CALL_LOOP_EX(full_listener_name,  output_iface, ::cpptesk::hw::HIGHEST_PRIORITY)

#ifdef HW_SHORT_NAMES
    #define DEFINE_BASIC_OUTPUT_LISTENER CPPTESK_DEFINE_BASIC_OUTPUT_LISTENER
    #define CALL_OUTPUT_LISTENER         CPPTESK_CALL_OUTPUT_LISTENER
#endif // HW_SHORT_NAMES

///////////////////////////////////////////////////////////////////////////////
// Capturing/Releasing Interfaces
///////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_CAPTURE_IFACE
 * Captures an interface.
 */
#define CPPTESK_CAPTURE_IFACE() \
    do { \
        CPPTESK_GET_IFACE().capture(); \
    } while(false)

/** @def CPPTESK_RELEASE_IFACE
 * Releases an interface.
 */
#define CPPTESK_RELEASE_IFACE() \
     do { \
         CPPTESK_GET_IFACE().release(); \
     } while(false)

#ifdef HW_SHORT_NAMES
    #define CAPTURE_IFACE() CPPTESK_CAPTURE_IFACE()
    #define RELEASE_IFACE() CPPTESK_RELEASE_IFACE()
#endif // HW_SHORT_NAMES

namespace cpptesk {
namespace hw {

}} // namespace cpptesk::hw

